/*
 * Copyright (c) 1997 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 * Copyright (c) 1997 Dug Song <dugsong@UMICH.EDU>
 *                    Nathan Binkert <binkertn@UMICH.EDU>
 *                    Jim Simpson <freyjs@UMICH.EDU>
 *
 * All rights reserved, all wrongs reversed.
 * Disclaimer: i claim dis
 *
 * Portions Copyright 1996 Netscape Communications.
 */

/* Standard includes. */
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "npapi.h"

/* Kerberos support. */
#include "npkrb4.h"

/* Plugin parameters. */
#define COOKIE_LIFETIME	5
#define PLUGIN_VERSION "1.3"

/* Unix cruft. */
#ifdef XP_UNIX
#define PLUGIN_NAME "Kerberos v4 Authentication Plug-in"
#define PLUGIN_DESCRIPTION "<a href=\"http://www-personal.umich.edu/~dugsong/kplugin/\">Kerberos v4 Authentication Plug-in</a> version 1.3"
#define PLUGIN_MIME_STRING "application/x-kerberos4-mime:krb4:Kerberos v4 Authentication"
#define PLUGIN_MIME_TYPE "application/x-kerberos4-mime"
#ifdef SOLARIS
#define snprintf __snprintf
#endif /* SOLARIS */
#endif /* XP_UNIX */

/* Mac cruft. */
#ifdef XP_MAC
#define rUserAlert 128
#ifdef __cplusplus
extern "C" 
{
#endif
int strncasecmp (const char *a, const char *b, int n);
void mac_time_to_unix_time (time_t *time);
#ifdef __cplusplus
}
#endif
CGrafPort gSavePort;
CGrafPtr gOldPort;
#endif /* XP_MAC */

/* Windows cruft. */
#ifdef XP_WINDOWS
LRESULT CALLBACK
PluginWindowProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam);
const char *gInstanceLookupString = "instance->pdata";
#ifdef XP_WIN32
#define snprintf _snprintf
#define strncasecmp _strnicmp
HANDLE TktCache;
#endif /* XP_WIN32 */
#endif /* XP_WINDOWS */

/* Instance definition. */
typedef struct _PluginInstance
{
#ifndef XP_UNIX
  NPWindow *fWindow;
  uint16 fMode;
#endif /* !XP_UNIX */
#ifdef XP_WINDOWS
  HWND fhWnd;
  WNDPROC fDefaultWindowProc;
#elif defined(XP_MAC)
  Handle fData;
#endif /* XP_WINDOWS || XP_MAC */
  
  C_Block session;	/* Kerberos session key */
  char ticket[1024];	/* Kerberos service ticket (base-64 encoded) */
  char buf[2048];	/* temp buffer */
} PluginInstance;

/* Our own alert dialog routine. */
void NPP_Alert(const char *fmt, ...)
{
  va_list args;
#ifndef XP_UNIX
  char buf[1024];
#endif /* !XP_UNIX */
  
  va_start(args, fmt);
#ifdef XP_WIN32
  snprintf(buf, sizeof(buf), fmt, args);
  MessageBox(NULL, buf, "Kerberos v4 authentication failed!", MB_OK);
#elif defined(XP_UNIX)
  fprintf(stderr, "Kerberos v4 authentication failed:\n");
  vfprintf(stderr, fmt, args);
  fprintf(stderr, "\n");
#elif defined(XP_MAC)
  vsprintf(buf, fmt, args);
  c2pstr(buf);
  ParamText((const unsigned char *)buf, NULL, NULL, NULL);
  StopAlert(rUserAlert, nil);
#endif /* XP_UNIX || XP_WIN32 || XP_MAC */
  va_end(args);
}

#ifdef XP_UNIX
char *NPP_GetMIMEDescription(void)
{
  return(PLUGIN_MIME_STRING);
}

NPError NPP_GetValue(void *future, NPPVariable variable, void *value)
{
  NPError err = NPERR_NO_ERROR;
  
  if (variable == NPPVpluginNameString)
    *((char **)value) = PLUGIN_NAME;
  else if (variable == NPPVpluginDescriptionString)
    *((char **)value) = PLUGIN_DESCRIPTION;
  else err = NPERR_GENERIC_ERROR;
  return err;
}
#endif /* XP_UNIX */

NPError NPP_Initialize(void)
{
#ifdef XP_WIN32
  /* Check for ticket cache and create one if it does not exist. NB */
  if (!KrbDoesTktCacheExist()) {
    if (!KrbCreateCache(50, &TktCache)) {
      MessageBox(NULL, "Could not create ticket cache", "Error", MB_OK);
      return NPERR_GENERIC_ERROR;
    }
    else TktCache = (HANDLE)0;
  }
#elif defined(XP_MAC)
  gSavePort.clipRgn = ::NewRgn();
#endif /* XP_WIN32 || XP_MAC */
  return NPERR_NO_ERROR;
}

jref NPP_GetJavaClass(void) { return NULL; }

void NPP_Shutdown(void)
{
#ifdef XP_WIN32
  /* Since I haven't figured out a way to pass ownership of the ticket
     cache, we must destroy it when we're done if we created it. NB */
  if (TktCache && !KrbReleaseCache(&TktCache))
    MessageBox(NULL, "Could not release ticket cache", "Error", MB_OK);
#elif defined(XP_MAC)
  if (gSavePort.clipRgn)
    ::DisposeRgn(gSavePort.clipRgn);
#endif /* XP_WIN32 || XP_MAC */
}

#define PUTLONG(l, cp) { \
	register unsigned int t_l = (unsigned int)(l); \
	register unsigned char *t_cp = (unsigned char *)(cp); \
	*t_cp++ = t_l >> 24; \
	*t_cp++ = t_l >> 16; \
	*t_cp++ = t_l >> 8; \
	*t_cp   = t_l; \
	(cp) += 4; }

int tkt_to_radix(KTEXT auth, char *buf)
{
  char *p, temp[1024];
  long len;
  
  p = temp;
  *p++ = 1; /* version */
  PUTLONG(auth->length, p);
  memcpy(p, auth->dat, auth->length);
  p += auth->length;
  len = p-temp;
  return uuencode((unsigned char *)temp, len, buf);
}

/* Get Kerberos credentials, get base-64 encoded ticket. */
NPError NPP_GetKrbAuth(NPP instance, const char *principal)
{
  PluginInstance *This;
#ifdef XP_UNIX
  CREDENTIALS cred;		/* Kerberos credentials */
  int i;
#elif defined(XP_WIN32)
  char prin[MAX_K_NAME_SZ];
  LPBYTE skey;
#elif defined(XP_MAC)
  _V4ticketInfo cred;
  short len, aref, avers = 0;
  int i, authman = 0;
  char *service_temp, *inst_temp, *realm_temp;
  OSErr err;
#endif /* XP_UNIX || XP_MAC || XP_WIN32 */

  KTEXT_ST auth;		/* Kerberos authenticator */
  char service[ANAME_SZ];	/* Kerberos service */
  char inst[INST_SZ];		/* Kerberos instance */
  char realm[REALM_SZ];		/* Kerberos realm */
  
  if (!instance || !principal)
    return NPERR_INVALID_INSTANCE_ERROR;
  This = (PluginInstance*) instance->pdata;
  
#ifdef XP_UNIX
  /* Parse principal into service, instance, realm for mk_req(). */
  *service = '\0'; *inst = '\0'; *realm = '\0';
  if (kname_parse(service, inst, realm, principal) != KSUCCESS ||
      !*service || !*inst || !*realm) {
    NPP_Alert("Incomplete or incorrect <EMBED> attributes.");
    return NPERR_INVALID_INSTANCE_ERROR;
  }
  /* Get service ticket. */
  if (i = krb_mk_req(&auth, service, inst, realm, 0L) != KSUCCESS) {
    NPP_Alert("Couldn't get service ticket for %s.%s@%s:\n%s",
	      service, inst, realm, krb_err_txt[i]);
    return NPERR_GENERIC_ERROR;
  }
  /* Save session key for use in generating HMAC later. */
  if (krb_get_cred(service, inst, realm, &cred) != KSUCCESS)
    return NPERR_GENERIC_ERROR;
  memcpy(This->session, cred.session, DES_KEY_SZ);
  
#elif defined(XP_MAC)
  /* Parse principal into service, instance, realm. */
  service_temp = strtok(principal, ".");
  inst_temp = strtok(NULL, "@");
  realm_temp = strtok(NULL, "\n");
  
  if (!service_temp || !inst_temp || !realm_temp) {
    NPP_Alert("Incomplete or incorrect <EMBED> attributes.");
    return NPERR_INVALID_INSTANCE_ERROR;
  }
  strncpy((char *)service, service_temp, sizeof(service));
  strncpy((char *)inst, inst_temp, sizeof(inst));
  strncpy((char *)realm, realm_temp, sizeof(realm));
    
  /* Convert these strings to \p strings, Authman/Mac specific annoyance(?) */
  c2pstr(service); c2pstr(realm); c2pstr(inst);
  
  authman = (openAuthMan(&aref, &avers) == noErr);
  if (authman && avers < 3) {
    closeAuthMan(aref);
    NPP_Alert("Authentication Manager is either too old, or not installed.");
    return NPERR_GENERIC_ERROR;
  }
  if (getV4Ticket(aref, (void *)auth.dat, &len,
                  (Str255 *)service, (Str255 *)inst, (Str255 *)realm,
		  60L * 5L /* 5 min */, PROMPT_FLAG)) {
    p2cstr((unsigned char*)service);
    p2cstr((unsigned char *)inst);
    p2cstr((unsigned char *)realm);
    NPP_Alert("Couldn't get service ticket for %s.%s@%s",
	      service, inst, realm);
    return NPERR_GENERIC_ERROR;
  }
  auth.length = len;
  
  for (i = 0; getV4TicketNinfo(aref, i, &cred) == noErr; ++i) {
    if (IUEqualString(cred.sName, (const unsigned char *)service) == 0 &&
	IUEqualString(cred.sInstance, (const unsigned char *)inst) == 0 &&
	IUEqualString(cred.sRealm, (const unsigned char *)realm) == 0) {
      break;
    }
  }
  /* Save session key for use in generating HMAC later. */
  memcpy(This->session, cred.sessionKey, DES_KEY_SZ);
  
#elif defined(XP_WIN32)
  strncpy(prin, principal, MAX_K_NAME_SZ-1);
  *service = '\0'; *inst = '\0'; *realm = '\0';
  KrbParseIdentity(prin, service, inst, realm);
  if (!*service || !*inst || !*realm) {
    NPP_Alert("Kerberos v4 authentication failed:\n"
	      "Incomplete or incorrect <EMBED> attributes.");
    return NPERR_INVALID_INSTANCE_ERROR;
  }
  if (!KrbBuildAuthorization(service, inst, realm, 0, 0,
			     &(auth.length), auth.dat, 4)) {
    NPP_Alert("Could not get authorization.");
    return NPERR_GENERIC_ERROR;
  }
  skey = KrbGetSessionKey(service, inst, realm, 4);
  if (!skey) return NPERR_GENERIC_ERROR;

  memcpy(This->session, skey, DES_KEY_SZ);

#endif /* XP_UNIX || XP_MAC || XP_WIN32 */

  /* Encode service ticket as RFC 1421 printable string. */
  if (tkt_to_radix(&auth, This->ticket) == 0) {
    NPP_Alert("Couldn't encode service ticket for %s.%s@%s",
	      service, inst, realm);
    return NPERR_GENERIC_ERROR;
  }
  return NPERR_NO_ERROR;
}

/* Generate HMAC-MD5 on URL directory. */
NPError make_hmac_md5(char *digest, const char *url, unsigned char *session)
{
  char *dir, tmp[1024];

  memset(digest, 0, 33); strncpy(tmp, url, sizeof(tmp));
  if (strlen(tmp) < 9) return NPERR_GENERIC_ERROR;
  if ((dir = strstr(tmp+9, "/")) == NULL) dir = "/";
  else if (dir[strlen(dir)] != '/') *(strrchr(dir, '/')+1) = '\0';
  hmac_md5((unsigned char *)dir, strlen(dir), session, DES_KEY_SZ, digest);
  return NPERR_NO_ERROR;
}
 
/* Kerberized version of NPN_GetURL(). */
NPError NPP_KrbGetURL(NPP instance, char *url,
	              const char *window, const char *principal)
{
  PluginInstance *This;
  char digest[33];
  const char *const days[7] =
  { "Sun","Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
  const char months[12][4] =
  {"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
  time_t when;
  struct tm *tms;
  
  if (!instance || !url || !window || !principal)
    return NPERR_INVALID_INSTANCE_ERROR;
  This = (PluginInstance*) instance->pdata;

  /* Get Kerberos service ticket. */
  if (NPP_GetKrbAuth(instance, principal) != NPERR_NO_ERROR)
    return NPERR_GENERIC_ERROR;
  
  /* Generate HMAC-MD5 of URL(+QUERY_STRING, if any). */
  make_hmac_md5(digest, url, (unsigned char *)This->session);
  
  /* Save our auth info in a cookie to be used for this request. */
  when = time(NULL)+COOKIE_LIFETIME;
  tms = gmtime(&when);
#ifdef XP_MAC
  mac_time_to_unix_time(&when);
  sprintf(This->buf,
#else
  snprintf(This->buf, sizeof(This->buf),
#endif /* XP_MAC */
	   "javascript:document.cookie="
	   "'KerberosV4=\"ticket=\"%s\"checksum=\"%s\"expires=\"%d\"version=\"%s\"\"; "
	   "expires=%s, %.2d-%s-%.2d %.2d:%.2d:%.2d GMT; path=/'; "
	   "location.reload()",
	   This->ticket, digest, when, PLUGIN_VERSION, days[tms->tm_wday], tms->tm_mday,
	   months[tms->tm_mon], (tms->tm_year >= 100) ? tms->tm_year - 100 :
	   tms->tm_year, tms->tm_hour, tms->tm_min, tms->tm_sec);
  
  /* Make authenticated request for page and display. */
  return NPN_GetURL(instance, This->buf, window);
}

/* Kerberized version of NPN_PostURLNotify(). Only takes buffer, no file. */

NPError NPP_KrbPostURLNotify(NPP instance, char *url, const char *window,
			     const int postlen, const char *postdata,
			     const char *principal, const char *notifyData)
{
  PluginInstance *This;
  char digest[33];
  time_t when;
  
  if (!instance || !url || !window || !postdata || !principal)
    return NPERR_INVALID_INSTANCE_ERROR;
  This = (PluginInstance*) instance->pdata;
  
  /* Get Kerberos service ticket. */
  if (NPP_GetKrbAuth(instance, principal) != NPERR_NO_ERROR)
    return NPERR_GENERIC_ERROR;
  
  /* Generate HMAC-MD5 of URL. */
  make_hmac_md5(digest, url, (unsigned char *)This->session);

  /* Save our auth info in a cookie to be used for this request. */
  when = time(NULL)+COOKIE_LIFETIME;
#ifdef XP_MAC
  mac_time_to_unix_time(&when);
  sprintf(This->buf,
#else
  snprintf(This->buf, sizeof(This->buf),
#endif /* XP_MAC */
	   "Cookie: KerberosV4=\"ticket=\"%s\"checksum=\"%s\""
	   "expires=\"%d\"version=\"%s\"\"\n"
	   "Content-type: application/x-www-form-urlencoded\n"
	   "Content-length: %d\n\n%s\n",
	   This->ticket, digest, when, PLUGIN_VERSION, postlen, postdata);

  NPN_PostURLNotify(instance, url, window, strlen(This->buf),
		    This->buf, FALSE, (void *)notifyData);
  
  return NPERR_NO_ERROR;
}

NPError NPP_New(NPMIMEType pluginType, NPP instance, uint16 mode,
		int16 argc, char* argn[], char* argv[], NPSavedData* saved)
{
  PluginInstance* This;
  char *principal, *url, *postdata;
  int i;
  
  if (instance == NULL)
    return NPERR_INVALID_INSTANCE_ERROR;
  
  /* Allocate memory for a new instance of our plug-in. */
  instance->pdata = NPN_MemAlloc(sizeof(PluginInstance));
  if (!(This = (PluginInstance*)instance->pdata)) {
    NPP_Alert("No available memory to load plug-in!");
    return NPERR_OUT_OF_MEMORY_ERROR;
  }
#ifdef XP_WIN32
  /* Mode is NP_EMBED, NP_FULL, or NP_BACKGROUND (see npapi.h). NB */
  This->fWindow = NULL;
  This->fMode = mode;
  This->fhWnd = NULL;
  This->fDefaultWindowProc = NULL;
#endif /* XP_WIN32 */
  
  /* Process arguments (attributes from EMBED tag). */
  principal = url = postdata = NULL;
  for (i = 0; i < argc; i++) {
    if (strncasecmp(argn[i], "PRINCIPAL", 9) == 0)
      principal = argv[i];
    if (strncasecmp(argn[i], "URL", 3) == 0)
      url = argv[i];
    if (strncasecmp(argn[i], "POST", 4) == 0)
      postdata = argv[i];
  }
  /* Check to see we got all required local tags. */
  if (!*url || !*principal) {
    NPP_Alert("Incomplete or incorrect <EMBED> attributes from server.");
    return NPERR_INVALID_INSTANCE_ERROR;
  }
  /* Get/post document with Kerberos authentication. */
  if (postdata)
    return NPP_KrbPostURLNotify(instance, url, "_current", strlen(postdata),
				postdata, principal, NULL);
  return NPP_KrbGetURL(instance, url, "_current", principal);
}

NPError NPP_Destroy(NPP instance, NPSavedData** save)
{
  PluginInstance* This;
  
  if (instance == NULL)
    return NPERR_INVALID_INSTANCE_ERROR;
  if (!(This = (PluginInstance*) instance->pdata))
    return NPERR_NO_ERROR;
#ifdef XP_MAC
  This->fWindow = NULL;
  This->fData = NULL;
#elif defined(XP_WINDOWS)
  /* If we have a window, clean it up. NB */
  if ( This->fWindow != NULL ) { 
    SetWindowLong( This->fhWnd, GWL_WNDPROC, (LONG)This->fDefaultWindowProc);
    This->fDefaultWindowProc = NULL;
    This->fhWnd = NULL;
  }
#endif /* XP_MAC || XP_WINDOWS */
  
  /* Free our instance. */
  NPN_MemFree(instance->pdata);
  instance->pdata = NULL;
  
  return NPERR_NO_ERROR;
}

NPError NPP_SetWindow(NPP instance, NPWindow* window)
{
  PluginInstance* This;
  
  if (instance == NULL) return NPERR_INVALID_INSTANCE_ERROR;
#ifdef XP_UNIX
  if (window == NULL) return NPERR_NO_ERROR;
#endif /* XP_UNIX */
  This = (PluginInstance*) instance->pdata;
  
#ifdef XP_MAC
  This->fWindow = window;
#elif defined(XP_WINDOWS)
  if (This->fWindow != NULL) {
    if ((window == NULL) || (window->window == NULL)) {
	  SetWindowLong(This->fhWnd, GWL_WNDPROC,
			(LONG)This->fDefaultWindowProc);
	  This->fDefaultWindowProc = NULL;
	  This->fhWnd = NULL;
	  This->fWindow=window;
	  return NPERR_NO_ERROR;
    }
    else if (This->fhWnd == (HWND) window->window) {
      InvalidateRect( This->fhWnd, NULL, TRUE );
      UpdateWindow( This->fhWnd );
      This->fWindow=window;
      return NPERR_NO_ERROR;
    }
    else {
      SetWindowLong(This->fhWnd, GWL_WNDPROC, (LONG)This->fDefaultWindowProc);
      This->fDefaultWindowProc = NULL;
      This->fhWnd = NULL;
    }
  }
  else if ((window == NULL) || (window->window == NULL)) {
    This->fWindow=window;
    return NPERR_NO_ERROR;
  }
  This->fDefaultWindowProc =
    (WNDPROC)SetWindowLong((HWND)window->window, GWL_WNDPROC,
			   (LONG)PluginWindowProc);
  This->fhWnd = (HWND) window->window;
  SetProp(This->fhWnd, gInstanceLookupString, (HANDLE)This);
  
  InvalidateRect(This->fhWnd, NULL, TRUE);
  UpdateWindow( This->fhWnd );
  
  This->fWindow = window;
#endif /* XP_MAC || XP_WINDOWS */
  
  return NPERR_NO_ERROR;
}

NPError NPP_NewStream(NPP instance, NPMIMEType type, NPStream *stream,
		      NPBool seekable, uint16 *stype)
{
/*  PluginInstance* This; */
  
  if (instance == NULL) return NPERR_INVALID_INSTANCE_ERROR;

  return NPERR_NO_ERROR;
}

int32 STREAMBUFSIZE = 0X0FFFFFFF; /* If we are reading from a file in NPAsFile
				   * mode so we can take any size stream in our
				   * write call (since we ignore it) */

int32 NPP_WriteReady(NPP instance, NPStream *stream) { return STREAMBUFSIZE; }


int32 NPP_Write(NPP instance, NPStream *stream, int32 offset,
		int32 len, void *buffer) { return len; }

NPError NPP_DestroyStream(NPP instance, NPStream *stream, NPError reason)
{ return NPERR_NO_ERROR; }

void NPP_StreamAsFile(NPP instance, NPStream *stream, const char* fname) {}

void NPP_Print(NPP instance, NPPrint* printInfo) {}

void NPP_URLNotify(NPP instance, const char* url, NPReason reason,
		   void* notifyData) {}

#ifdef XP_MAC
int16 NPP_HandleEvent(NPP instance, void* event)
{
  Boolean eventHandled = false;
  if (instance == NULL) return eventHandled;
  
  PluginInstance* This = (PluginInstance*) instance->pdata;
  if (This != NULL && event != NULL) {
    EventRecord* ev = (EventRecord*) event;
    switch (ev->what) {
    case updateEvt:
      eventHandled = true;
      break;
    default:
      break;
    }
  }
  return eventHandled;
}
#elif defined(XP_WIN32)
int16 NPP_HandleEvent(NPP instance, void* event) { return 0; }

LRESULT CALLBACK PluginWindowProc(HWND hWnd, UINT Msg, WPARAM wParam,
				  LPARAM lParam)
{
  PluginInstance *This =
    (PluginInstance*)GetProp(hWnd, gInstanceLookupString);
  
  switch(Msg) {
  case WM_PAINT:
    {
      PAINTSTRUCT paintStruct;
      HDC hdc;
      
      hdc = BeginPaint(hWnd, &paintStruct);
      TextOut(hdc, 0, 0, "", 0);
      
      EndPaint( hWnd, &paintStruct );
      break;
    }
  default:
    This->fDefaultWindowProc( hWnd, Msg, wParam, lParam);
  }
  return 0;
}
#endif /* XP_MAC || XP_WIN32 */

/* EOF */
